# =====================================
# generator=datazen
# version=3.1.3
# hash=10ebf2374e8ea826d1191aa1df9ef925
# =====================================

"""
Useful defaults and other package metadata.
"""

DESCRIPTION = "A network-application TUI using textual."
PKG_NAME = "conntextual"
VERSION = "1.2.0"



"""
An entry-point for the 'ui' command.
"""

# built-in
from argparse import ArgumentParser as _ArgumentParser
from argparse import Namespace as _Namespace

# third-party
from runtimepy.commands.common import arbiter_args
from runtimepy.entry import main as runtimepy_main
from vcorelib.args import CommandFunction as _CommandFunction
from vcorelib.logging import forward_flags

# internal
from conntextual import PKG_NAME

DEFAULT_VARIANT = "app"


def ui_cmd(args: _Namespace) -> int:
    """Execute the ui command."""

    cli_args = ["runtimepy"]

    flags = set(forward_flags(args, ["curses", "verbose", "no_uvloop"]))

    # Don't initialize regular logging no matter what.
    flags.add("--quiet")

    cli_args.extend(flags)

    cli_args.append("arbiter")
    cli_args.extend(list(forward_flags(args, ["init_only"])))

    # Always wait for the program to exit.
    cli_args.append("-w")

    cli_args.extend(args.configs)
    cli_args.append(f"package://{args.package}/{args.variant}.yaml")

    print(f"runtimepy_main({cli_args})")
    return runtimepy_main(cli_args)


def add_ui_cmd(parser: _ArgumentParser) -> _CommandFunction:
    """Add ui-command arguments to its parser."""

    parser.add_argument(
        "-p",
        "--package",
        default=PKG_NAME,
        help="package to source application from",
    )
    parser.add_argument(
        "-v",
        "--variant",
        choices=[DEFAULT_VARIANT, "curses"],
        default=DEFAULT_VARIANT,
        help="application variant to use (default: %(default)s)",
    )

    arbiter_args(parser, nargs="*")

    return ui_cmd



"""
A module implementing a plot widget.
"""

# third-party
from numpy.typing import ArrayLike
import plotext as plt
from rich.text import Text
from textual.widgets import Static


class Plot(Static):
    """A plot widget."""

    theme = "pro"

    def __init__(
        self,
        x: ArrayLike,
        y: ArrayLike,
        *args,
        title: str = "under construction",
        **kwargs,
    ) -> None:
        """Initialize this instance."""

        super().__init__(*args, **kwargs)
        self.x = x
        self.y = y
        self.title = title

    def on_show(self) -> None:
        """Handle showing the plot."""

        self.dispatch()

    def on_resize(self) -> None:
        """Handle re-size."""

        self.dispatch()

    def dispatch(self) -> None:
        """Draw a new instance of the plot."""

        plt.clf()

        plt.plot(self.x, self.y)
        plt.title(self.title)
        plt.theme(self.theme)

        plt.plotsize(self.size.width, self.size.height)

        self.update(Text.from_ansi(plt.build()))

    def set_data(self, x: ArrayLike, y: ArrayLike) -> None:
        """Assign new data."""

        self.x = x
        self.y = y
        self.dispatch()



"""
A module implementing user interface elements for channel environments.
"""

# built-in
import random
from typing import Dict, List, Tuple, Union

# third-party
from rich.text import Text
from runtimepy.channel import AnyChannel
from runtimepy.channel.environment import ChannelEnvironment
from runtimepy.net.arbiter import AppInfo
from textual import on
from textual.app import ComposeResult
from textual.containers import HorizontalScroll, ScrollableContainer
from textual.coordinate import Coordinate
from textual.widgets import DataTable, Pretty, Static
from vcorelib.logging import LoggerType
from vcorelib.math import to_nanos

# internal
from conntextual.ui.channel.color import type_str_style
from conntextual.ui.channel.log import ChannelEnvironmentLog
from conntextual.ui.channel.model import ChannelEnvironmentSource, Model
from conntextual.ui.channel.plot import Plot
from conntextual.ui.channel.selected import SelectedChannel
from conntextual.ui.channel.suggester import CommandSuggester

__all__ = ["ChannelEnvironmentDisplay"]
COLUMNS = ["id", "type", "name", "value"]
DEFAULT_VALUE_COL_WIDTH = 22
STALE_THRESHOLD_NS = to_nanos(0.5)


def css_name(name: str) -> str:
    """Replace some characters that don't work in identifier values."""
    return name.replace(".", "_")


class ChannelEnvironmentDisplay(Static):
    """A channel-environment interface element."""

    model: Model

    by_index: List[Tuple[Coordinate, AnyChannel]]
    channels_by_row: Dict[int, SelectedChannel]

    selected: SelectedChannel

    def on_mount(self) -> None:
        """Populate channel table."""

        table = self.query_one(DataTable)
        env = self.model.env
        assert env.finalized
        names = list(env.names)

        # Set up columns.
        table.add_columns(*COLUMNS)
        value_column: int = COLUMNS.index("value")

        row_idx = 0

        for name in names:
            chan, enum = env[name]

            kind_str = str(chan.type)

            # Should handle enums at some point.
            if enum is not None:
                enum_name = env.enums.names.name(enum.id)
                assert enum_name is not None
                kind_str = enum_name

            table.add_row(
                chan.id,
                Text(kind_str, style=type_str_style(chan.type, enum)),
                name
                if not chan.commandable
                else Text(name, style="bold green"),
                " "
                * max(len(str(env.value(chan.id))), DEFAULT_VALUE_COL_WIDTH),
            )
            self.by_index.append((Coordinate(row_idx, value_column), chan))
            self.channels_by_row[row_idx] = SelectedChannel.create(
                name, (chan, enum)
            )
            row_idx += 1

    def switch_to_channel(self, row: int) -> None:
        """Switch the plot to a channel at the specified row."""

        # Select channel.
        self.selected = self.channels_by_row[row]

        # Update plot parameters.
        name = self.selected.name
        self.query_one(Plot).title = name
        self.model.logger.info("Switched plot to channel '%s'.", name)
        self.reset_plot()

    def random_channel(self) -> None:
        """Switch to a random channel."""

        self.switch_to_channel(
            random.choice(list(self.channels_by_row.keys()))
        )

    def reset_plot(self) -> None:
        """Reset the selected plot."""

        self.selected.reset()
        self.query_one(Plot).set_data(
            self.selected.timestamps, self.selected.values
        )
        self.model.logger.info("Plot reset.")

    @on(DataTable.CellSelected)
    def handle_cell_selected(self, event: DataTable.CellSelected) -> None:
        """Handle input submission."""

        self.switch_to_channel(event.coordinate.row)

    def update_channels(self) -> None:
        """Update all channel values."""

        env = self.model.env
        table = self.query_one(DataTable)

        for coord, chan in self.by_index:
            val = env.value(chan.id)
            if isinstance(val, float):
                val = f"{val: 15.6f}"
            elif isinstance(val, bool):
                val = "true" if val else "false"
            elif isinstance(val, int):
                val = f"{val: 8d}       "

            # Get the age of the primitive.
            age = env[chan.id][0].raw.age_ns()
            if age > STALE_THRESHOLD_NS:
                val = Text(val, style="yellow")  # type: ignore

            table.update_cell_at(coord, val)

        # Update logs.
        self.query_one(ChannelEnvironmentLog).dispatch()

        # Update plot.
        self.selected.poll()
        self.query_one(Plot).dispatch()

    @property
    def label(self) -> str:
        """Obtain a label string for this instance."""
        return f"({self.model.source}) {self.model.name}"

    def compose(self) -> ComposeResult:
        """Create child nodes."""

        with HorizontalScroll(classes="channels"):
            yield DataTable[Union[str, int, float]]()

            yield Plot(
                self.selected.timestamps,
                self.selected.values,
                title=self.selected.name,
                id="plot",
            )

        # Create log and command widget.
        log = ChannelEnvironmentLog()
        log.parent_name = self.model.name
        log.logger = self.model.logger
        log.suggester = CommandSuggester.create(self.model.env, log.logger)
        yield log

        with ScrollableContainer():
            yield Pretty(self.model.app.config.get("root", {}))

    @staticmethod
    def create(
        name: str,
        env: ChannelEnvironment,
        source: ChannelEnvironmentSource,
        logger: LoggerType,
        app: AppInfo,
    ) -> "ChannelEnvironmentDisplay":
        """Create a channel-environment display."""

        result = ChannelEnvironmentDisplay(id=css_name(name))
        result.model = Model(name, env, source, logger, app)
        result.by_index = []
        result.channels_by_row = {}

        name = random.choice(list(env.names))
        result.selected = SelectedChannel.create(name, env[name])

        return result
